         
#include <Arduino.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <SPIFFS.h>
#include <ESPAsyncWebServer.h>
#include "driver/i2s.h"
#include "freertos/queue.h"

#include "Types/Type_B.h"
#include "Types/Type_J.h"
#include "Types/Type_K.h"
#include "Types/Type_N.h"
#include "Types/Type_R.h"
#include "Types/Type_S.h"
#include "Types/Type_T.h"

AsyncWebServer server(80);

// REPLACE WITH YOUR NETWORK CREDENTIALS
const char* ssid = "REPLACE_WITH_YOUR_SSID";
const char* password = "REPLACE_WITH_YOUR_PASSWORD";

// Переменные вывода
float termo_table; // пересчитанное значение из градуировочной таблицы, mV, со знаком 
uint16_t termo_uV; //пересчитанное значение, uV, без знака,
uint16_t termo; // оно же инвертированное 
uint32_t termo_out; // 32 разрядный вывод

const int zero = 32768; // сдвиг выходного напряжения

const char* PARAM_STRING = "inputString";
const char* PARAM_INT = "inputInt";
const char* PARAM_INT2 = "inputInt2"; 

const char* PARAM_INPUT_1 = "output";
const char* PARAM_INPUT_2 = "state";

String Type; // Тип термопары
uint16_t Cold; // температура холодного спая
uint16_t Hot; // температура горячего спая

// PWM параметры для ИП негативного напряжения
const int ledPin = 27;  
const int freq = 128000;
const int ledChannel = 0;
const int resolution = 4;
const int dutyCycle = 8;

static const i2s_port_t i2s_num = I2S_NUM_0; // i2s port number
    
// HTML web page to handle 3 input fields (inputString, inputInt, inputInt2)
const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html><head>
  <title>ESP Input Form</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <meta http-equiv="Content-Type" content="text/html; charset=utf-8">
    <style>
    html {font-family: Arial; display: inline-block; text-align: center;}
    h2 {font-size: 3.0rem;}
    p {font-size: 3.0rem;}
    body {max-width: 600px; margin:0px auto; padding-bottom: 25px;}
    .switch {position: relative; display: inline-block; width: 120px; height: 68px} 
    .switch input {display: none}
    .slider {position: absolute; top: 0; left: 0; right: 0; bottom: 0; background-color: #ccc; border-radius: 6px}
    .slider:before {position: absolute; content: ""; height: 52px; width: 52px; left: 8px; bottom: 8px; background-color: #fff; -webkit-transition: .4s; transition: .4s; border-radius: 3px}
    input:checked+.slider {background-color: #b30000}
    input:checked+.slider:before {-webkit-transform: translateX(52px); -ms-transform: translateX(52px); transform: translateX(52px)}
  </style>
</head>
<body>
  <h4>Симулятор термопар</h4>
  %BUTTONPLACEHOLDER%
  <script>
    function submitMessage() {
      alert("Saved value to ESP SPIFFS");
      setTimeout(function(){ document.location.reload(false); }, 500);   
    }
  </script></head><body>
 
  <script>function toggleCheckbox(element) {
  var xhr = new XMLHttpRequest();
  if(element.checked){ xhr.open("GET", "/update?output="+element.id+"&state=1", true); }
  else { xhr.open("GET", "/update?output="+element.id+"&state=0", true); }
  xhr.send();
}
</script>
  <form action="/get" target="hidden-form">
  <h4>  Тип - B,J,K,N,R,S,T (сейчас <font style="color: Red">%inputString%</font>): <input type="text" name="inputString"> </h4>
  <input type="submit" value="Submit" onclick="submitMessage()">
  </form><br>
  <form action="/get" target="hidden-form">
  <h4> Холодный спай, &#8451; (сейчас <font style="color: Red">%inputInt%</font>): <input type="number " name="inputInt"> </h4>
    <input type="submit" value="Submit" onclick="submitMessage()">
  </form><br>
  <form action="/get" target="hidden-form">
  <h4> Горячий спай, &#8451; (сейчас <font style="color: Red">%inputInt2%</font>): <input type="number " name="inputInt2"> </h4>
    <input type="submit" value="Submit" onclick="submitMessage()">
  </form>
  <iframe style="display:none" name="hidden-form"></iframe>
</body></html>)rawliteral";

void notFound(AsyncWebServerRequest *request) {
  request->send(404, "text/plain", "Not found");
}

String readFile(fs::FS &fs, const char * path){
  //Serial.printf("Reading file: %s\r\n", path);
  File file = fs.open(path, "r");
  if(!file || file.isDirectory()){
    //Serial.println("- empty file or failed to open file");
    return String();
  }
  //Serial.println("- read from file:");
  String fileContent;
  while(file.available()){
    fileContent+=String((char)file.read());
  }
  file.close();
  //Serial.println(fileContent);
  return fileContent;
}

void writeFile(fs::FS &fs, const char * path, const char * message){
  //Serial.printf("Writing file: %s\r\n", path);
  File file = fs.open(path, "w");
  if(!file){
    //Serial.println("- failed to open file for writing");
    return;
  }
  if(file.print(message)){
    //Serial.println("- file written");
  } else {
    //Serial.println("- write failed");
  }
  file.close();
}

// Replaces placeholder with stored values
String processor(const String& var){
  //Serial.println(var);
 if(var == "BUTTONPLACEHOLDER"){
    String buttons = "";
    buttons += "<h4>Настройка <---> Работа</h4><label class=\"switch\"><input type=\"checkbox\" onchange=\"toggleCheckbox(this)\" id=\"2\" " + outputState(2) + "><span class=\"slider\"></span></label>";
    return buttons;
  }
  if(var == "inputString"){
    return readFile(SPIFFS, "/inputString.txt");
  }
  else if(var == "inputInt"){
    return readFile(SPIFFS, "/inputInt.txt");
  }
  else if(var == "inputInt2"){
    return readFile(SPIFFS, "/inputInt2.txt");
  }
  return String();
}

String outputState(int output){
  if(digitalRead(output)){
    return "checked";
  }
  else {
    return "";
  }
}


// I2S configuration structures

static const i2s_config_t i2s_config = {
    .mode = (i2s_mode_t)(I2S_MODE_MASTER | I2S_MODE_TX),
    .sample_rate = 2500,
    .bits_per_sample = I2S_BITS_PER_SAMPLE_16BIT,
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
    .communication_format = (i2s_comm_format_t)(I2S_COMM_FORMAT_I2S | I2S_COMM_FORMAT_I2S_MSB),
    .intr_alloc_flags = ESP_INTR_FLAG_LEVEL1,       // high interrupt priority
    .dma_buf_count = 8,                             // 8 buffers
    .dma_buf_len = 1024,                            // 1K per buffer, so 8K of buffer space
    .use_apll=0,
    .tx_desc_auto_clear= true, 
    .fixed_mclk=-1    
};

static const i2s_pin_config_t pin_config = {
    .bck_io_num = 26,                                 // The bit clock connectiom, goes to pin 27 of ESP32
    .ws_io_num = 25,                                  // Word select, also known as word select or left right clock
    .data_out_num = 22,                               // Data out from the ESP32, connect to DIN on 38357A
    .data_in_num = I2S_PIN_NO_CHANGE                  // we are not interested in I2S data into the ESP32
};

         
void setup() {
  // Заявляем цифровые входы и выходы
  pinMode(2, OUTPUT); // индикатор режима работы
  digitalWrite(2,LOW); // гасим синий светодиод - это режим Настройка 

  // PWM для источника негативного питания
  ledcSetup(ledChannel, freq, resolution);
  ledcAttachPin(ledPin, ledChannel);
     
  Serial.begin(115200);
  // Initialize SPIFFS
  if(!SPIFFS.begin(true)){
  Serial.println("An Error has occurred while mounting SPIFFS");
  return;
  }
  
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  if (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Serial.println("WiFi Failed!");
    return;
  }
  Serial.println();
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());

  // Send web page with input fields to client
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html, processor);
  });

  // Send a GET request to <ESP_IP>/get?inputString=<inputMessage>
  server.on("/get", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage;
    // GET inputString value on <ESP_IP>/get?inputString=<inputMessage>
    if (request->hasParam(PARAM_STRING)) {
      inputMessage = request->getParam(PARAM_STRING)->value();
      writeFile(SPIFFS, "/inputString.txt", inputMessage.c_str());
    }
    // GET inputInt value on <ESP_IP>/get?inputInt=<inputMessage>
    else if (request->hasParam(PARAM_INT)) {
      inputMessage = request->getParam(PARAM_INT)->value();
      writeFile(SPIFFS, "/inputInt.txt", inputMessage.c_str());
    }
    // GET inputInt2 value on <ESP_IP>/get?inputInt2=<inputMessage>
    else if (request->hasParam(PARAM_INT2)) {
      inputMessage = request->getParam(PARAM_INT2)->value();
      writeFile(SPIFFS, "/inputInt2.txt", inputMessage.c_str());
    }
    else {
      inputMessage = "No message sent";
    }
    //Serial.println(inputMessage);
    request->send(200, "text/text", inputMessage);
  });

   // Send a GET request to <ESP_IP>/update?output=<inputMessage1>&state=<inputMessage2>
  server.on("/update", HTTP_GET, [] (AsyncWebServerRequest *request) {
    String inputMessage1;
    String inputMessage2;
    // GET input1 value on <ESP_IP>/update?output=<inputMessage1>&state=<inputMessage2>
    if (request->hasParam(PARAM_INPUT_1) && request->hasParam(PARAM_INPUT_2)) {
      inputMessage1 = request->getParam(PARAM_INPUT_1)->value();
      inputMessage2 = request->getParam(PARAM_INPUT_2)->value();
      digitalWrite(inputMessage1.toInt(), inputMessage2.toInt());
    }
    else {
      inputMessage1 = "No message sent";
      inputMessage2 = "No message sent";
    }
    //Serial.print("GPIO: ");
    //Serial.print(inputMessage1);
    //Serial.print(" - Set to: ");
    //Serial.println(inputMessage2);
    request->send(200, "text/plain", "OK");
  });
  server.onNotFound(notFound);
  server.begin();
    
    ledcWrite(ledChannel, dutyCycle); // генератор для ИП негативного питания
    i2s_driver_install(i2s_num, &i2s_config, 0, NULL);        // ESP32 will allocated resources to run I2S
    i2s_set_pin(i2s_num, &pin_config);                        // Tell it the pins you will be using                                            
}

void loop(){    
     
    if (digitalRead(2) == LOW) {
    Type = readFile(SPIFFS, "/inputString.txt"); // читаем тип термопары
    Cold = readFile(SPIFFS, "/inputInt.txt").toInt(); // читаем температуру холодного спая
    Hot = readFile(SPIFFS, "/inputInt2.txt").toInt(); // читаем температуру горячего спая 
 
    if (Type == "B") termo_table = Type_B[0   - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "J") termo_table = Type_J[200 - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "K") termo_table = Type_K[260 - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "N") termo_table = Type_N[260 - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "R") termo_table = Type_R[40  - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "S") termo_table = Type_S[40  - Cold + Hot]; // результат вычисления напряжения в mV 
    if (Type == "T") termo_table = Type_T[260 - Cold + Hot]; // результат вычисления напряжения в mV 
 
    termo_uV = termo_table * 1000; // пересчитанное значение, uV, без знака,
    termo = ~ (termo_uV - zero); // оно же инвертированное
     }
    else {
        if (termo_table < 0) digitalWrite(2, LOW); // если это напряжение отрицательное, то возращаемся в режим настройки 
        termo_out = (termo<<16) | (termo & 0xffff); // Output same value on both left and right channels 
        size_t BytesWritten;
        i2s_write(i2s_num,&termo_out,4,&BytesWritten,portMAX_DELAY ); // вывод в I2S   
      }
  
}
